<?php
/**
 * WooCommerce - Bizuno API
 * This class contains the common methods for interfacing with Bizuno, specifics are handled in the first included class
 *
 * @copyright 2008-2019, PhreeSoft, www.PhreeSoft.com
 * @author David Premo, PhreeSoft, Inc.
 * @version 3.x Last Update: 2019-07-31
 * bizunoAPI - This class handles the mapping functions for the downloading operations, the methods are cart specific
 * bizunoCommon - Common methods used for easier internal code generation and commonality amongst third party apps
 * bizunoPDO - Database class
 * @filesource wp-content/plugins/bizuno/bizuno-lib.php
 */

namespace bizunoapi;

if (!defined('SCRIPT_START_TIME')) { define('SCRIPT_START_TIME', microtime(true)); }
if (!defined('DEBUG'))             { define('DEBUG', true); }

if (!class_exists('bizunoAPI')):
class bizunoAPI extends bizunoCommon
{
    // Local variables that need to be set
    var $user_email = ''; // will be filled in after access granted

    function __construct() {
        global $wpdb;
        $GLOBALS['dbBizuno'] = ['type'=>'mysql','host'=>DB_HOST,'name'=>DB_NAME,'user'=>DB_USER,'pass'=>DB_PASSWORD,'prefix'=>$wpdb->prefix];
        $this->path_to_this_module = dirname(__FILE__).'/';
        define('DB_PREFIX', $GLOBALS['dbBizuno']['prefix']);
        parent::__construct($GLOBALS['dbBizuno']);
    }

    private function configGet($const='') {
        if (is_array($const)) {
            $output = [];
            foreach ($const as $value) {
                $output[$value] = $this->dbGetValue(DB_PREFIX."options", 'option_value', "option_name LIKE '$value'", true);
            }
        } else {
            $output = $const ? $this->dbGetValue(DB_PREFIX."options", 'option_value', "option_name LIKE '$const'", true) : '';
        }
        return $output;
    }

    /**
     * processes any response received by WooCommerce
     */
    public function processRcv() {
        if (get_magic_quotes_gpc()) { // if magic quotes is turned on, fail processing
            $this->msgAdd("Magic quotes is turned on please turn them off on your website! I cannot process your request.", 'error');
        } elseif ($this->validateUser()) {
            switch (sanitize_user($_POST['Action'], true)) {
                case 'uploadProduct': $this->uploadProduct(); break;
                case 'syncProduct'  : $this->syncProduct();   break;
                case 'shipConfirm'  : $this->shipConfirm();   break;
                case 'testCon'      : $this->testConection(); break;
                default : $this->msgAdd("error bad action!"); break;
            }
        } else { // check for WooCommerce webhook
            $this->webHook();
        }
        $this->msgOutput();
    }

    /**
     *
     * @param string $action
     * @param string $params
     * @return array with two elements 'result' => string message level 'error','warning','success'; 'message' => result message
     */
    public function processSend($action=false, $params=false, $data=false)
    {
        switch ($action) {
            case 'downloadOrder': // $params should contain only the order_id
                $url = $this->configGet('bizuno_api_url');
                $oID = is_numeric($params) ? intval($params) : 0;
                if (!$oID) { $this->msgAdd("Bad order ID!"); break; }
                if (!$map = $this->mapOrder($oID, $data)) { break; }
                if (!$url) { // local Bizuno
                    // check to make sure Bizuno is installed and active
                    require_once (ABSPATH.'wp-admin/includes/plugin.php');
                    if ( !is_plugin_active('bizuno-accounting/bizuno-accounting.php') ) {
                        $this->msgAdd("The bizuno plugin is not active on this site. The API will not work without it or enter a URL for a remote Bizuno site to connect with.");
                        break;
                    }
                    $this->msgDebug("Sending to local Bizuno with data: ".print_r($map, true));
                    $_POST = $map;
                    $_GET['p'] = 'bizuno/api/processRcv';
                    unset($_COOKIE['bizunoSession']); // in case user has been logged into wordpress, this forces new login with API creds
                    require_once (BIZUNO_API_PATH."../bizuno-accounting/portal/controller.php");
                    $ctl = new \bizuno\portalCtl();
                    $ctl->compose();
                    global $msgStack;
//                  $this->msgDebug("\nReturned from local compose with msg = ". print_r($msgStack->error, true));
                    $this->msgDebug("\n\nBizuno Accounting trace contents: \n\n".$msgStack->trace); // full trace
                    $output = $msgStack->error;
                    unset($_COOKIE['bizunoSession']); // sign off as API user, should reload cache with user's creds
                } else {
                    $path   = rtrim($url, '/')."/index.php?p=bizuno/api/processRcv";
                    $this->msgDebug("Sending to url = $path with data: ".print_r($map, true));
                    $results= wp_remote_post($path, ['method'=>'POST','timeout'=>45,'body'=>$map]);
                    $this->msgDebug("\nResponse error raw data = ".print_r($results, true));
                    $output = json_decode($results['body'], true);
                    if (json_last_error() <> JSON_ERROR_NONE) {
                        $this->msgAdd("Unexpected response from server: ".print_r($results['body'], true));
                    }
                }
                if (array_key_exists('message', (array)$output)) { $output = $output['message']; }
                $result = $this->setNotices(array_merge($this->arrMsg, (array)$output));
                if (in_array($result, ['success','warning'])) { update_post_meta($oID, 'bizuno_order_exported', 1); }
                break;
            default:
        }
        if (DEBUG) { $this->msgDebugWrite(); }
    }

    /**
     * Pulls the responses from the message stack and sets the WordPress notices for the next reload
     * @param array $msgStack
     */
    private function setNotices($msgStack=[])
    {
        $error = $warning = $success = "";
        if (empty($msgStack)) {
            \bizuno_api_add_notice("Unexpected response from the Bizuno API: ".print_r($msgStack, true), 'error', true);
            return;
        }
        foreach ($msgStack as $key => $value) {
            switch ($key) {
                case 'error':   foreach ($value as $msg) { $error   .= $msg['text']."\n"; } break;
                case 'caution':
                case 'warning': foreach ($value as $msg) { $warning .= $msg['text']."\n"; } break;
                case 'success': foreach ($value as $msg) { $success .= $msg['text']."\n"; } break;
            }
        }
        $this->msgDebug("\nOrder Download Error: $error"."\nOrder Download Warning: $warning"."\nOrder Download Success: $success");
        $result = 'error';
        if (!empty($success)) { $result = 'success'; \bizuno_api_add_notice( $success, 'success', true ); }
        if (!empty($warning)) { $result = 'warning'; \bizuno_api_add_notice( $warning, 'warning', true ); }
        if (!empty($error))   { $result = 'error';   \bizuno_api_add_notice( $error,   'error',   true ); }
        return $result;
    }

    /**
     * @todo IS THIS CALLED? auto download is working without using this function!
     * @return type
     */
    private function webHook()
    {
        $rawpost = json_decode(file_get_contents("php://input"), true);
        // some predefined tests
        if (isset($rawpost['order']['order_number'])) { // WooCommerce Webhook for AutoOrder Download
            $this->arrMsg = array(); // reset the messageStack, has validation error flagged
            $this->processSend('downloadOrder', $rawpost['order']['order_number']);
            if (!isset($this->arrMsg['error'])) { // no errors set the downloaded flag
                update_post_meta((int)$rawpost['order']['order_number'], 'bizuno_order_exported', 1);
            }
        }
        return $this->arrMsg;
    }

    /**
    *
    * @param unknown_type $data
    * @return void|Ambigous
    */
    private function mapOrder($order_id, $data=array()) {
        $settings = $this->configGet(array('bizuno_api_user', 'bizuno_api_pw', 'bizuno_api_comp', 'bizuno_api_prefix', 'bizuno_api_prefix_cust'));
        if (!isset($settings['bizuno_api_user']) || $settings['bizuno_api_user'] == '' || !isset($settings['bizuno_api_pw']) || $settings['bizuno_api_pw'] == '') {
            return $this->msgAdd("please enter user and password in woocommerce account info.");
        }
        // start pulling the data
        $order_post = $this->dbGetRow(DB_PREFIX."posts", "`id`='$order_id' AND `post_type` LIKE 'shop_order'");
        $order_post_meta = $this->dbGetMulti(DB_PREFIX."postmeta", "post_id='$order_id'", '', ['meta_key','meta_value']);
        foreach ($order_post_meta as $meta_num ) { $order_info[$meta_num['meta_key']] = $meta_num['meta_value']; }
        $order_item = $this->dbGetMulti(DB_PREFIX."woocommerce_order_items", "`order_id`=$order_id AND `order_item_type` LIKE 'line_item'");
        foreach ($order_item as $item_info) {
            $item_meta = $this->dbGetMulti(DB_PREFIX."woocommerce_order_itemmeta", "order_item_id = {$item_info['order_item_id']}", '', ['meta_key' , 'meta_value']);
            foreach ($item_meta as $row) { $item_formated[$row['meta_key']] = $row['meta_value']; }
            $order_item_meta[$item_info['order_item_name']] = $item_formated;
        } //this is a array ([order_item_id] => array ( [meta_key] => [meta_value] ))
        //error_log("order meta: ".json_encode(print_r($item_formated,true)));
        // Get the Shipping method, this may move depending on shipping plug-in used
        $shipping_method = $this->dbGetValue(DB_PREFIX."woocommerce_order_items", "order_item_name", "order_id=$order_id AND order_item_type='shipping'");
        $customer_id = $order_info['_billing_email'];
        $total_tax = $order_info['_order_tax']+$order_info['_order_shipping_tax']+$order_info['_cart_discount_tax'];
        $this->msgDebug("\nMapping order = ".print_r($order_info, true));
        $Map = [
            'General' => [
                'OrderID'         => $settings['bizuno_api_prefix'] . $order_id,
                'PurchaseOrderID' => $settings['bizuno_api_prefix'] . $order_id,
                'OrderDate'       => substr($order_post['post_date'], 0, strpos($order_post['post_date'], ' ')),
                'OrderTotal'      => $order_info['_order_total'],
                'SalesTaxAmount'  => $total_tax,
//                'DiscountTotal'   => $this->fetchTotal('coupon', $totals_query->rows),
//                'SalesTaxPercent' => '',
//                'SalesTaxTitle'   => '',
                'ShippingTotal'   => $order_info['_order_shipping'],
                'ShippingCarrier' => $shipping_method, //$this->shipmentMap($order_info['_shipping_method']),
                'OrderNotes'      => $order_post['post_excerpt']],
            'Payment' => [
                'Method'          => $order_info['_payment_method'],
                'Title'           => !empty($order_info['_payment_method_title']) ? $order_info['_payment_method_title'] : $order_info['_payment_method'],
                'Status'          => $this->guessPmtStatus($order_info), // $order_info['_payment_cap_auth'],
                'Authorization'   => $order_info['_payment_auth_code'], // Authorization code from credit cards that need to be captured to complete the sale
                'TransactionID'   => $order_info['_payment_transaction_id'], // transaction ID from gateway
                'Hint'            => $order_info['_payment_card_hint']], // credit card hint, format 12**********7890
            'Billing' => [
                'CustomerID'      => $customer_id,
                'CompanyName'     => $order_info['_billing_company'] != '' ? $order_info['_billing_company'] : $order_info['_billing_first_name'].' '.$order_info ['_billing_last_name'],
                'Contact'         => $order_info['_billing_company'] == '' ? $order_info['_billing_first_name'].' '.$order_info['_billing_last_name'] : '',
                'Address1'        => $order_info['_billing_address_1'],
                'Address2'        => $order_info['_billing_address_2'],
                'City'            => $order_info['_billing_city'],
                'State'           => $order_info['_billing_state'],
                'PostalCode'      => $order_info['_billing_postcode'],
                'Country'         => $this->get_country_iso($order_info['_billing_country']),
                'Telephone'       => $order_info['_billing_phone'],
                'Email'           => $order_info['_billing_email']],
            'Shipping' => [
                'CompanyName'     => $order_info['_shipping_company'] != '' ? $order_info['_shipping_company'] : $order_info['_shipping_first_name'].' '.$order_info ['_shipping_last_name'],
                'Contact'         => $order_info['_shipping_company'] == '' ? $order_info['_shipping_first_name'].' '.$order_info['_shipping_last_name'] : '',
                'Address1'        => $order_info['_shipping_address_1'],
                'Address2'        => $order_info['_shipping_address_2'],
                'City'            => $order_info['_shipping_city'],
                'State'           => $order_info['_shipping_state'],
                'PostalCode'      => $order_info['_shipping_postcode'],
                'Country'         => $this->get_country_iso($order_info['_shipping_country']),
                'Telephone'       => $order_info['_billing_phone'],
                'Email'           => $order_info['_billing_email']],
            ];
        $Map['Item'] = [];
        foreach ($order_item_meta as $item_name => $item_info) {
        //    $item_tax_total = $item['product_quantity'] * $item['tax'];
            $Map['Item'][] = [
                'ItemID'          => $this->fetchSKU($item_info['_product_id'], $item_info['order_item_name']),
                'Description'     => $item_name,
                'Quantity'        => $item_info['_qty'],
                'UnitPrice'       => $item_info['_line_subtotal']/$item_info['_qty'],
                'SalesTaxPercent' => '',
                'SalesTaxTitle'   => '',
                'SalesTaxAmount'  => $item_info['_line_tax'],
                'TotalPrice'      => $item_info['_line_total']];
        //    $Map['General']['TaxOrder'] -= $item_tax_total;
        }
        $Map['Other'] = $data;
        return [
            'UserID'  => $settings['bizuno_api_user'],
            'UserPW'  => $settings['bizuno_api_pw'],
            'Version' => '2.00',
            'Action'  => 'journal',
            'Order'   => json_encode($Map, JSON_UNESCAPED_UNICODE)];
    }

    /**
     * Attempts to guess the payment status for the order, either cap (Sale) or auth (Authorize Only), defaults to cap
     * @param array $order_info - Order details
     * @return string - either cap or auth depending on the order payment status
     */
    private function guessPmtStatus($order_info) {
        switch ($order_info['_payment_method']) {
            default:
            case 'elavon': return isset($order_info['_payment_cap_auth']) ? $order_info['_payment_cap_auth'] : 'cap';
            case 'paypal': // works for WooCommerce included PayPal Standard payment method
                return isset($order_info['_paypal_status']) && $order_info['_paypal_status']=='completed' ? 'cap' : 'auth';
        }
    }

    /**
     * @param unknown_type $id
     * @param unknown_type $name
     * @return Ambigous <>|unknown
     */
    private function fetchSKU($id, $name) {
        $result = $this->dbGetValue(DB_PREFIX."postmeta", "meta_value","post_id=$id AND meta_key LIKE '_sku'", true);
        if ($result !== false) {
            return $result;
        }
        return $name;
    }

    /**
     * This validates the User information and checks for the right information.
     *
     * WordPress validation at wp-includes/pluggable.php -> wp_check_password (1937)
     * @param String $username
     * @param String $password
     * @return boolean
     */
    public function validateUser() {
        $username = sanitize_email($_POST['UserID']);
        $password = sanitize_text_field($_POST['UserPW']);
        if (!$username || !$password) { return $this->msgAdd("Failed validation, user= $username and pw = $password, could not detect username/password!"); }
        //
        require_once(ABSPATH.WPINC.'/class-phpass.php');
        $wp_hasher = new \PasswordHash(8, true);
        $row = $this->dbGetRow(DB_PREFIX.'users',"user_login='$username' OR user_email='$username'");
        if (!$row) { return $this->msgAdd('Wordpress: Cannot find account in your wordpress site.'); }
        $check = $wp_hasher->CheckPassword($password, $row['user_pass']);
        if (!$check) { return $this->msgAdd("Failed validation, invalid username/password combination!"); }
        // Now check role to make sure they have the permission to connect
        $user_lvl = $this->dbGetValue(DB_PREFIX.'usermeta', 'meta_value', "user_id={$row['ID']} AND meta_key='".DB_PREFIX."user_level'");
        $this->msgDebug("User ID {$row['ID']} with level = $user_lvl");
        if ($user_lvl < 6) { return $this->msgAdd("You do not have sufficient permission to acess this API!"); }
        $this->user_email = $this->dbGetValue(DB_PREFIX.'users', 'user_email', "ID LIKE '{$row['ID']}'");
        return true;
    }

    /**
     * Looks for product and replaces or creates a new product
     * @param array $bizuno_array - data array of the product information.
     * @return null - messages are added with status
     */
    public function uploadProduct() {
        $product = json_decode(stripslashes_deep($_POST['Product']), true);
        $language= substr(sanitize_key($_POST['Language']), 0, 2);
        if (empty($product))  { $this->msgAdd("Bad json string received, check the error log."); }
        if (empty($language)) { $language = 'en'; }
        $this->msgDebug("\nWorking with product: ".print_r($product, true));
        if (!$this->productValidate($product, $language)) { return false; }
        $this->productImport($product);
    }

    private function productValidate(&$product, $language) {
        // error check input
        if (!isset($product['SKU'])) { return $this->msgAdd("Cannot find product SKU."); }
        // fetch related id
        if (isset($product['invAccessory'])) {
            $product['related'] = array();
            foreach ($product['invAccessory'] as $related) {
                $product_id = $this->dbGetValue(DB_PREFIX.'postmeta', 'post_id', "`meta_key` LIKE '_sku' AND `meta_value`='{$related}'", true);
                if ($product_id !== false) { $product['related'][] = $product_id; }
            }
            $this->msgDebug("related items found:".print_r($product['related'], true));
        }
        return true;
    }

    private function productImport($product) {
        $user_id = get_current_user_id(); // was $this->dbGetValue(DB_PREFIX.'users', 'ID', "user_login LIKE '$username'");
        // build the products table data
        $product_post_array = [
            'post_author'          => $user_id, //id
            'post_date'            => !empty($product['DateCreated'])     ? $product['DateCreated'] : date('Y-m-d H:i:s'),
            'post_date_gmt'        => !empty($product['DateCreated'])     ? $product['DateCreated'] : date('Y-m-d H:i:s'),
            'post_content'         => !empty($product['DescriptionLong']) ? $product['DescriptionLong'] : $product['DescriptionSales'],
            'post_title'           => $product['Description'],
            'post_excerpt'         => !empty($product['DescriptionSales'])? $product['DescriptionSales'] : $product['Description'],
            'post_status'          => 'publish', // allows the product to be displayed (default)
            'post_name'            => $this->getPermaLink($product['Description']),
            'comment_status'       => 'open',
            'ping_status'          => 'closed',
            'post_password'        => '',
            'to_ping'              => '',
            'pinged'               => '',
            'post_modified'        => !empty($product['DateLastUpdate'])  ? $product['DateLastUpdate'] : date('Y-m-d H:i:s'),
            'post_modified_gmt'    => !empty($product['DateLastUpdate'])  ? $product['DateLastUpdate'] : date('Y-m-d H:i:s'),
            'post_content_filtered'=> '',
            'post_parent'          => 0,//post catagory
//          'guid'                 => '', // need to set after ID is known
            'post_type'            => 'product', //,
            'post_mime_type'       => ''];
        $product_meta_array = [
            '_downloadable'        => !empty($product['Downloadable'])   ? 'yes': 'no',
            '_manage_stock'        => !empty($product['Virtual'])        ? 'no' : 'yes',
            '_price'               => !empty($product['SalePrice'])      ? $product['SalePrice']   : (!empty($product['RegularPrice']) ? $product['RegularPrice'] : $product['Price']),
            '_regular_price'       => !empty($product['RegularPrice'])   ? $product['RegularPrice']: $product['Price'],
            '_sku'                 => $product['SKU'],
            '_stock'               => $product['QtyStock'] > 0 ? $product['QtyStock'] : 0,
            '_stock_status'        => $product['QtyStock'] > 0 ? 'instock' : 'outofstock',
            '_tax_class'           => '',
            '_tax_status'          => 'taxable',
            '_backorders'          => !empty($product['Backorder'])      ? $product['Backorder'] : 'notify', // Options: 'no', 'yes', 'notify'
            '_upsell_ids'          => '', // serialized array
            '_virtual'             => !empty($product['Virtual'])        ? 'yes' : 'no',
            '_visibility'          => 'visible'];
        if (!empty($product['SalePrice'])) { $product_meta_array['_sale_price'] = $product['SalePrice']; } // only change if sent in feed, otherwise managed @ Woo
        //covert weight (need to convert kg,lb,oz,g)
        if (isset($product['WeightUOM'])) {
            $weight = strtolower($product['WeightUOM']);
            $wordpress_weight = strtolower($this->configGet('woocommerce_weight_unit'));
            $product_meta_array['_weight'] = isset($product['Weight']) ? $this->convertWeight($product['Weight'], $dim, $wordpress_dim) : '';
        }
        //covert dim (need to convert m,cm,mm,in,yd)
        if (isset($product['DimensionUOM'])) {
            $dim = strtolower($product['DimensionUOM']);
            $wordpress_dim = strtolower($this->configGet('woocommerce_dimension_unit'));
            $product_meta_array['_length']= isset($product['ProductLength'])? $this->convertLength($product['ProductLength'],$dim, $wordpress_dim) : '';
            $product_meta_array['_width'] = isset($product['ProductWidth']) ? $this->convertLength($product['ProductWidth'], $dim, $wordpress_dim) : '';
            $product_meta_array['_height']= isset($product['ProductHeight'])? $this->convertLength($product['ProductHeight'],$dim, $wordpress_dim) : '';
        }
        //post the product
        $productID = $this->getProductID($product['SKU']);
        $new_product = $productID ? false : true;
        $product_id = $this->setPost($product_post_array, $product_meta_array, $productID);
        if ($new_product) { // need some new termmeta data
            $product_meta_array = [
                '_edit_last'            => '',
                'total_sales'           => '0',
                '_edit_lock'            => '',
                '_purchase_note'        => '',
                '_featured'             => 'no',
                '_product_attributes'   => 'a:():{}',
                '_sale_price'           => '',
                '_sale_price_dates_from'=> '',
                '_sale_price_dates_to'  => '',
                '_sold_individually'    => '',
                '_crosssell_ids'        => 'a:():{}',
                '_product_version'      => $this->getVersion(),
                '_product_image_gallery'=> ''];
            $product_id = $this->setPost([], $product_meta_array, $product_id);
        }
        // update link to product
        $this->dbWrite(DB_PREFIX."posts", ['guid'=>$_SERVER['HTTP_HOST'].'/?post_type=product&#038;p='.$product_id], 'update', "ID=$product_id");
        // update the price levels
        $this->productPrices($product, $product_id);
        //update category
        $this->productCategory($product, $product_id);
        // Update attributes
        $this->productAttributes($product, $product_id);
        //update related
/*        if (isset($product['related'])) {
            $this->dbGetResult("DELETE FROM `". DB_PREFIX . "postmeta` WHERE post_id = '". (int)$product_id . "' AND meta_key = '_crosssell_ids';");
            $this->dbGetResult("INSERT INTO " . DB_PREFIX . "postmeta SET post_id = '"   . (int)$product_id . "', meta_key = '_crosssell_ids' , meta_value = '" . $product['related'] . "';");
        }
*/
        // Set images
        $this->productImage($product, $product_id, $user_id);
        // update the category counts in the term_taxonomy table
        $this->dbGetResult("UPDATE ".DB_PREFIX."term_taxonomy SET count = (SELECT COUNT(*) FROM ".DB_PREFIX."term_relationships rel LEFT JOIN ".DB_PREFIX."posts po ON (po.ID = rel.object_id)
            WHERE rel.term_taxonomy_id = ".DB_PREFIX."term_taxonomy.term_taxonomy_id AND ".DB_PREFIX."term_taxonomy.taxonomy NOT IN ('link_category') AND po.post_status IN ('publish', 'future'))");
        $this->msgAdd("Successfully imported SKU: {$product['SKU']}", 'success');
        return $product_id;
    }

    private function getProductID($sku) {
        $candidates = $this->dbGetMulti(DB_PREFIX.'postmeta', "`meta_value`='$sku' AND `meta_key`='_sku'", '', ['post_id']);
        foreach ($candidates as $val) {
            $postID = $this->dbGetValue(DB_PREFIX.'posts', 'ID', "`ID`='{$val['post_id']}' AND `post_status`='publish'");
            if ($postID) { return $postID; }
        }
        return 0;
    }

    /**
     * Enters the volume price if included with feed
     *
     * REQUIRES THE WooCommerce Bulk Pricing plug-in set to percentage discount (default)
     * https://wordpress.org/plugins/woocommerce-bulk-discount/
     *
     * @param type $product
     * @param type $product_id
     * @return type
     */
    private function productPrices($product, $product_id)
    {
        $resetMeta = true;
        if (!empty($product['PriceLevels']) && is_array($product['PriceLevels']) ) {
            foreach ($product['PriceLevels'] as $price) {
                if (empty($price['default'])) { continue; }
                $description = '<table style="border:1px solid black;width:300px"><tr><th colspan="2">Quantity Pricing</th></tr><tr><th>Qty</th><th>Price</th></tr>';
                foreach ($price['levels'] as $idx => $level) {
                    if (!$idx) { // first price is single qty, save to calc discount
                        $full_price = !empty($level['price']) ? $level['price'] : 999999;
                        continue;
                    } // skip the first price level as it is the single unit price
                    $resetMeta = false; // if we are here, there is more than 1 price in this discount group
                    $description .= '<tr><td style="border: 1px solid black;text-align:center">'.$level['qty'].'</td><td style="border: 1px solid black;text-align:center">$ '.number_format($level['price'], 2).'</td></tr>';
                    $price_meta_array['_bulkdiscount_quantity_'.($idx)] = $level['qty'];
                    // uncomment for Percentage Discount, both are allowed
                    $price_meta_array['_bulkdiscount_discount_'.($idx)] = ((1 - $level['price']/$full_price) * 100); // percentage
                    // uncomment for Fixed Discount
                    $price_meta_array['_bulkdiscount_discount_fixed_'.($idx)] = $full_price - $level['price']; // fixed discount
                }
            }
            $description .= '</table>';
        }
        if ($resetMeta) { // clear any fields that may have been there from a prior discount
            $this->dbGetResult("DELETE FROM ".DB_PREFIX."postmeta WHERE post_id=$product_id AND meta_key LIKE '_bulkdiscount_quantity_%'");
            $this->dbGetResult("DELETE FROM ".DB_PREFIX."postmeta WHERE post_id=$product_id AND meta_key LIKE '_bulkdiscount_discount_%'"); // also does _bulkdiscount_discount_fixed_%
            $this->dbGetResult("DELETE FROM ".DB_PREFIX."postmeta WHERE post_id=$product_id AND meta_key LIKE '_bulkdiscount_text_info'");
            $price_meta_array['_bulkdiscount_enabled']  = 'no';
        } else {
            $price_meta_array['_bulkdiscount_enabled']  = 'yes';
            $price_meta_array['_bulkdiscount_text_info']= $description;
        }
        $this->setPost([], $price_meta_array, $product_id);
    }

    private function productCategory($product, $product_id)
    {
        // Parse the category
        if (!isset($product['WooCommerceCategory']) || !$product['WooCommerceCategory']) {
            return $this->msgAdd("Error - the category was not passed for product: {$product['SKU']}, it must be set manually in WooCommerce.", 'caution');
        }
        $categories = explode(";", $product['WooCommerceCategory']);
        $parent = 0;
        foreach ($categories as $category) {
            $breadcrumbs = explode(":", $category);
            foreach ($breadcrumbs as $breadcrumb) { // starts at the top level and goes down.
                if (empty($breadcrumb)) { continue; }
                $term_id = $this->dbGetValue(DB_PREFIX.'terms t JOIN '.DB_PREFIX.'term_taxonomy x ON t.term_id=x.term_id', "term_taxonomy_id",
                        "x.taxonomy='product_cat' AND x.parent=$parent AND t.name='$breadcrumb'", false);
                if (empty($term_id)) { // not found, create the term
                    $my_cat = [
                        'cat_name'            => $breadcrumb,
                        'category_description'=> str_replace(':', ' ', $category),
                        'category_nicename'   => strtolower(str_replace([':',' '], '-', $category)),
                        'category_parent'     => $parent,
                        'taxonomy'            => 'product_cat'];
                    $term_id = wp_insert_category($my_cat);
//                  $this->msgAdd("Info - the category: $category was not found, it has been created, parent = $parent returned id = $term_id with params = ".print_r($my_cat, true), 'caution');
                }
                // Check every box down the category tree
                $rID = $this->dbGetValue(DB_PREFIX.'term_relationships', 'object_id', "object_id=$product_id AND term_taxonomy_id=$term_id");
                if (!$rID) { $this->dbWrite(DB_PREFIX.'term_relationships', ['object_id'=>$product_id,'term_taxonomy_id'=>$term_id]); }
                $parent = $term_id;
            }
            // Check just the lowest on the categeory tree
//          $rID = $this->dbGetValue(DB_PREFIX.'term_relationships', 'object_id', "object_id=$product_id AND term_taxonomy_id=$term_id");
//          if (!$rID) { $this->dbWrite(DB_PREFIX.'term_relationships', ['object_id'=>$product_id,'term_taxonomy_id'=>$term_id]); }
        }
        return true;
    }

    private function productAttributes($product, $product_id)
    {
        if (!isset($product['Attributes']) || !sizeof($product['Attributes'])) { return; }
        $result = $this->dbGetMulti(DB_PREFIX.'term_taxonomy', "taxonomy LIKE 'pa_%'");
        $pa_attr_ids = [];
        foreach ($result as $row) { $pa_attr_ids[] = $row['term_taxonomy_id']; }
        if (sizeof($pa_attr_ids)) { // clear out the current attributes
            $this->dbGetResult("DELETE FROM ".DB_PREFIX."term_relationships WHERE object_id=$product_id AND term_taxonomy_id IN (".implode(',',$pa_attr_ids).")");
        }
        $productAttr = [];
        foreach ($product['Attributes'] as $idx => $row) {
            if (empty($row['title'])) { continue; }
            $attribute_name = $this->dbGetValue(DB_PREFIX.'woocommerce_attribute_taxonomies', 'attribute_name', "attribute_label='{$row['title']}'");
            if (!$attribute_name) {
                $attribute_name = $this->getPermaLink($row['title']);
                $newAttr = [
                    'attribute_name'   => $attribute_name,
                    'attribute_label'  => $row['title'],
                    'attribute_type'   => 'text',
                    'attribute_orderby'=> 'name_num',
                    'attribute_public' => 0];
                $this->dbWrite(DB_PREFIX.'woocommerce_attribute_taxonomies', $newAttr);
            }
            // see if value taxonomy exists
            $valSlug = $this->getPermaLink($row['value']);
            $termSlug = "pa_$attribute_name-$valSlug";
            $term_id = $this->dbGetValue(DB_PREFIX.'terms', 'term_id', "slug='$termSlug'");
            if (!$term_id) {
                $term_id    = $this->dbWrite(DB_PREFIX.'terms',        ['name'=>$row['value'],'slug'=>$termSlug]);
                $meta_id    = $this->dbWrite(DB_PREFIX.'termmeta',     ['term_id'=>$term_id,'meta_key'=>"order_pa_$attribute_name",'meta_value'=>$idx+1]);
                $taxonomy_id= $this->dbWrite(DB_PREFIX.'term_taxonomy',['term_id'=>$term_id,'taxonomy'=>"pa_$attribute_name"]);
            }
            if (isset($taxonomy_id) || !$taxonomy_id) {
                $taxonomy_id = $this->dbGetValue(DB_PREFIX.'term_taxonomy', 'term_taxonomy_id', "term_id=$term_id");
            }
            $term_relationship = $this->dbGetValue(DB_PREFIX.'term_relationships', 'object_id', "object_id=$product_id AND term_taxonomy_id=$taxonomy_id");
            if (!$term_relationship) {
                $this->dbWrite(DB_PREFIX.'term_relationships', ['object_id'=>$product_id,'term_taxonomy_id'=>$taxonomy_id]);
            }
            $productAttr["pa_$attribute_name"] = ["name"=>"pa_$attribute_name","value"=>"","position"=>$idx,"is_visible"=>1,"is_variation"=>0,"is_taxonomy"=>1];
        }
        $this->dbWrite(DB_PREFIX.'postmeta', ['meta_value'=>serialize($productAttr)], 'update', "post_id=$product_id AND meta_key='_product_attributes'");
    }

    /**
     *
     * @return type
     */
    private function getVersion()
    {
        return $this->dbGetValue(DB_PREFIX.'options', 'option_value', "option_name='woocommerce_version'");
    }

    /**
     *
     * @param type $value
     * @return type
     */
    private function getPermaLink($value)
    {
        $test1 = str_replace(" ", "-", trim($value));
        $test2 = preg_replace("/[^a-zA-Z0-9_-]/", "", $test1);
        while (strpos($test2, '--') !== false) { $test2 = str_replace('--', '-', $test2); }
        return strtolower($test2);
    }

    private function productImage($product, $product_id, $user_id=0) {
        $this->msgDebug("\nEntering Image function with product ID = $product_id");
        $path    = 'products/'.(!empty($product['ProductImageDirectory']) ? $product['ProductImageDirectory'] : ''); // from root upload folder
        $mainImg = [
            'path' => $path . (substr($path, -1)<>"/" ? "/" : ''),
            'name' => isset($product['ProductImageFilename'])? $product['ProductImageFilename']: '',
            'data' => isset($product['ProductImageData'])    ? $product['ProductImageData']    : ''];
        $imgMain = $this->dbGetMulti(DB_PREFIX.'postmeta', "post_id=$product_id AND meta_key='_thumbnail_id'");
        if (sizeof($imgMain) > 1) { // clean up duplicate main images by deleting all and creating a new one
            $this->dbGetResult("DELETE FROM ".DB_PREFIX."postmeta WHERE post_id=$product_id AND meta_key='_thumbnail_id'");
            foreach ($imgMain as $row) { \wp_delete_post( $row['meta_value'], true ); }
            $imgID = 0;
        } else {
            $imgID = sizeof($imgMain)==1 ? $imgMain[0]['meta_value'] : 0;
        }
        $mID  = $this->setImage($mainImg, $imgID, $user_id);
        $this->dbWrite(DB_PREFIX.'postmeta', ['post_id'=>$product_id, 'meta_key'=>'_thumbnail_id', 'meta_value'=>$mID], $imgID?'update':'insert', "meta_id=$imgID");
        if (empty($product['Images'])) { return; }// new style multiple images allowed
        $metaID = $this->dbGetValue(DB_PREFIX.'postmeta', 'meta_id', "post_id=$product_id AND meta_key='_product_image_gallery'");
        foreach ($product['Images'] as $image) {
            $path    = 'products/'.(!empty($image['Path']) ? $image['Path'] : ''); // from root upload folder
            $images[]= ['path'=>$path.(substr($path, -1)<>"/" ? "/" : ''), 'name'=>$image['Filename'], 'data'=>$image['Data']];
        }
        $xIDs = [];
        foreach ($images as $image) {
            $imgID  = $this->dbGetValue(DB_PREFIX.'posts', 'ID', "post_name='{$image['name']}' AND post_type='attachment' AND post_mime_type LIKE 'image/%'");
            $xIDs[] = $this->setImage($image, $imgID, $user_id);

        }
        $meta  = ['post_id'=>$product_id, 'meta_key'=>'_product_image_gallery', 'meta_value'=>implode(',', $xIDs)];
        $this->dbWrite(DB_PREFIX.'postmeta', $meta, $metaID?'update':'insert', "meta_id=$metaID");
    }

    /**
     * Uploads the image and puts it in the proper folder, creates to folder if it can.
     * @param array $image
     * @param integer $imgID
     * @param integer $user_id
     * @return type
     */
    private function setImage($image, $imgID=0, $user_id=0) {
        $upload_folder  = wp_upload_dir(); // was "wp-content/uploads/"; // need trailing slash
        $image_directory= $upload_folder['basedir']."/{$image['path']}";
        $this->msgDebug("\nTesting for image length = ".strlen($image['data']));
        if (!$image['data']) { return; } // no image was sent up to save, just return with no message
        // NOTE: the str_replace is to necessary to fix a PHP 5 issue with spaces in the base64 encode... see php.net
        $contents    = base64_decode(str_replace(" ", "+", $image['data']));
        if (!is_dir($image_directory)) { if (!@mkdir($image_directory, 0755, true)) { return $this->msgAdd("Cannot create image path: $image_directory"); } }
        $full_path   = $image_directory.$image['name'];
        $dirname     = dirname($full_path);
        if (!is_dir($dirname)) { mkdir($dirname, 0755, true); }
        if (!$handle = fopen($full_path, 'wb')) { return $this->msgAdd("Cannot open Image path: $full_path"); }
        if (fwrite($handle, $contents) === false) { return $this->msgAdd("Cannot write Image file."); }
        fclose($handle);
        $this->msgDebug("\nWrote image image_directory = $image_directory and image_filename = {$image['name']} and image length = ".strlen($image['data']));
        $postData = [
            'post_author'          => $user_id, //id
            'post_date'            => date('Y-m-d H:i:s'),
            'post_date_gmt'        => date('Y-m-d H:i:s'),
            'post_content'         => '', // was $product['DescriptionSales'] but is not used so remove it
            'post_title'           => $image['name'], // was $product['Description'], but needs to be filename without the extension
            'post_excerpt'         => '', // was isset($product['DescriptionSales'])? $product['DescriptionSales']: $product['Description'], but is not used so remove it
            'post_status'          => 'inherit', // allows the product to be displayed (default)
            'comment_status'       => 'closed',
            'ping_status'          => 'closed',
            'post_password'        => '',//no password
            'post_name'            => $image['name'], // was 'img-'.$this->getPermaLink($product['Description']), // needs a counter if multiple images, should use filename
            'post_modified'        => date('Y-m-d H:i:s'),
            'post_modified_gmt'    => date('Y-m-d H:i:s'),
            'post_content_filtered'=> '',
            'post_parent'          => 0, //$product_id,//post catagory
            'guid'                 => $upload_folder['baseurl']."/{$image['path']}".$image['name'],
            'post_type'            => 'attachment',
            'post_mime_type'       => $this->getImageType($image['name'])];
        $info = getimagesize($full_path);
        $metaData = [
            'width'     => $info[0],
            'height'    => $info[1],
            'file'      => "{$image['path']}{$image['name']}",
            'sizes'     => [],
            'image_meta'=> []];
        $postMeta = [
            '_wp_attached_file'       => $image['path'].$image['name'],
            '_wp_attachment_metadata' => serialize($metaData),
//          '_wp_attachment_image_alt'=> isset($product['DescriptionSales']) ? $product['DescriptionSales'] : $product['Description'],
            ];
        return $this->setPost($postData, $postMeta, $imgID);
    }

    private function getImageType($entry)
    {
        $ext = strtolower(substr($entry, strrpos($entry, '.')+1));
        if (in_array($ext, ['png'])) { return 'image/png'; }
        if (in_array($ext, ['jpg','jpeg'])) { return 'image/jpeg'; }
        if (in_array($ext, ['gif'])) { return 'image/gif'; }
        return '';
    }

    /**
     * This method syncs the products flagged in Bizuno to be listed and the actual listed products.
     * If syncDelete flag is set, the products in the cart will be deleted if they are not on the Bizuno list
     * @return messageStack entries
     */
    public function syncProduct()
    {
        $bizunoProducts= json_decode(stripslashes_deep($_POST['SKUs']), true);
        $cartProducts  = $this->dbGetMulti(DB_PREFIX."postmeta", "meta_key='_sku'", 'meta_key');
        $missing_skus  = array();
        foreach ($cartProducts as $product_db) {
            if (!in_array($product_db['meta_value'], $bizunoProducts['Product'])) {
                $title = $this->dbGetValue(DB_PREFIX."posts", 'post_title', "ID='{$product_db['post_id']}'");
                $missing_skus[] = $title.' (SKU: '.$product_db['meta_value'].')';
                if ($bizunoProducts['syncDelete'] == '1') {
                    $product_id = (int)$product_db['post_id'];
                    $this->dbGetResult("DELETE FROM ".DB_PREFIX."posts WHERE ID=$product_id");
                    $this->dbGetResult("DELETE FROM ".DB_PREFIX."postmeta WHERE post_id=$product_id");
                }
            }
        }
        if (sizeof($missing_skus) > 0) {
            $this->msgAdd($this->lang['msg_sku_missing'].'Bizuno:'.'<br>'.implode('<br>', $missing_skus), 'caution');
        } else {
            $this->msgAdd($this->lang['msg_sku_sync_success'], 'success');
        }
    }

    /**
     * Updates order status
     * @return null - fills the meessageStack
     */
    public function shipConfirm() {

        $orders    = json_decode(stripslashes_deep($_POST['Order']), true);
        $order_cnt = 0;
        $order_list= [];
        $prefix    = $this->configGet('bizuno_api_prefix');
        $status    = $this->configGet('bizuno_ship_status');
        $newStatus = $status ? $status : 'wc-completed';
        foreach ($orders as $id => $value) {
            $this->msgDebug("\nWorking with id = $id and value = ".print_r($value, true));
            // strip prefix from order ref (WC3044 => 3044) that is the record number
            if     ($prefix && (strpos($id, $prefix) !== 0 || strpos($id, $prefix) === false)) { continue; }
            elseif ($prefix && strpos($id, $prefix) === 0) { $id = substr($id, strlen($prefix)); }
            if (intval($id) != $id) { continue; }
            // get the post_status, comment_status, comment_count from the order,
            $result = $this->dbGetValue(DB_PREFIX.'posts', ['post_status','comment_status','comment_count'], "ID='$id'");
            // if not found OR wp-completed continue - no need to duplicate
            if (!$result || $result['post_status'] == $newStatus) { $this->msgDebug("\nSkipping id=$id"); continue; }
            // change the status to (wp-completed, closed, bump comment_count)
            $data = ['post_status'=>$newStatus,'comment_status'=>'closed','comment_count' =>$result['comment_count']+1];
            $this->dbWrite(DB_PREFIX.'posts', $data, 'update', "ID='$id'");
            // insert into wp_comments per table fields with tracking, ship date etc.
            $comment = [
                'comment_post_ID'     => $id,
                'comment_author'      => 'Bizuno_Interfce',
                'comment_author_email'=> $this->user_email,
                'comment_date'        => date('Y-m-d H:i:s'),
                'comment_date_gmt'    => gmdate('Y-m-d H:i:s'),
                'comment_content'     => $value['Tracking'],
                'comment_agent'       => 'WooCommerce',
                'comment_type'        => 'order_note',
            ];
            $this->dbWrite(DB_PREFIX.'comments', $comment);
            $order_cnt++;
            $order_list[] = $id;
        }
        $this->msgAdd(sprintf($this->lang['confirm_success'], sizeof($order_list), sizeof($order_list)>0?" (".implode(', ', $order_list).")":''), 'success');
    }

    private function testConection()
    {
        $this->msgAdd($this->lang['msg_test_success'], 'success');
    }

    private function setPost($postData, $postMeta, $post_id=0)
    {
        if (sizeof($postData) > 0) {
            $id = $this->dbWrite(DB_PREFIX.'posts', $postData, $post_id?'update':'insert', "ID=$post_id");
            if (!$post_id) { $post_id = $id; }
        }
        // read the postmeta
        $current = $this->dbGetMulti(DB_PREFIX.'postmeta', "post_id=$post_id");
        // put into keyed array
        $working = [];
        foreach ($current as $row) { $working[$row['meta_key']] = $row['meta_id']; }
        foreach ($postMeta as $meta_key => $meta_data) { //update meta data
            $rID = isset($working[$meta_key]) ? $working[$meta_key] : false;
            $data = ['post_id'=>$post_id, 'meta_key'=>$meta_key, 'meta_value'=>$meta_data];
            $this->dbWrite(DB_PREFIX.'postmeta', $data, $rID?'update':'insert',"meta_id=$rID");
        }
        return $post_id;
    }

    /**
     *
     * @param unknown $value
     * @param unknown $unit_original m,cm,mm,in,ft
     * @param unknown $unit_return m,cm,mm,in,yd
     */
    private function convertLength($value, $unit_original, $unit_return) {
        if($unit_original == $unit_return) { return $value; }
        switch ($unit_original) {
            case 'm':
                switch($unit_return) {
                    case 'cm': return $value/100;
                    case 'mm': return $value/1000;
                    case 'in': return $value*39.370;
                    case 'yd': return $value/1.0936;
                }
            case 'cm':
                switch($unit_return) {
                    case  'm': return $value*100;
                    case 'mm': return $value/10;
                    case 'in': return $value*0.39370;
                    case 'yd': return $value/109.36;
                }
            case 'mm':
                switch($unit_return) {
                    case  'm': return $value*1000;
                    case 'cm': return $value*10;
                    case 'in': return $value*0.039370;
                    case 'yd': return $value*0.039370*3*12;
                }
            case 'in':
                switch($unit_return) {
                    case  'm': return $value/39.370;
                    case 'cm': return $value/0.39370;
                    case 'mm': return $value/0.039370;
                    case 'yd': return $value*3*12;
                }
            case 'ft':
                switch($unit_return) {
                    case  'm': return $value/3.2808;
                    case 'cm': return $value/0.032808;
                    case 'mm': return $value/0.0032808;
                    case 'yd': return $value*3;
                    case 'in': return $value/12;
                }
            default:
                $this->msgAdd("length conversion Error","warning");
                return $value;
        }
    }

    /**
     *
     * @param unknown $value
     * @param unknown $unit_original kg,lb
     * @param unknown $unit_return kg,g,lb,oz
     * @return void|unknown|string|number
     */
    private function convertWeight($value, $unit_original, $unit_return) {
        if($unit_original == $unit_return) { return $value; }
        switch ($unit_original) {
            case 'kg':
                switch($unit_return) {
                    case 'g': return $value/1000;
                    case 'lb': return $value*2.2046;
                    case 'oz': return $value*35.274;
                }
            case 'lb':
                switch($unit_return) {
                    case 'kg': return $value/2.2046;
                    case 'g': return $value/0.0022046;
                    case 'oz': return $value*16;
                }
            default:
                $this->msgAdd("weight conversion Error","warning");
                return $value;
        }
    }
}
endif;

/**
 * This class contains common methods used across all import and export API's
 */
class bizunoCommon {
    // Bizuno common variables (generally do not need to touch)
    var $arrMsg = [];
    var $trap   = false; // when set to true, writes a debug trace file
    var $lang   = array(
        'err_io_file_open'    => 'Cannot open file (%s)',
        'err_io_file_write'   => 'Cannot write file (%s)',
        'sync_cart_missing'   => "Missing at the cart:",
        'sync_bizuno_missing' => "Products flagged to sync with PrestaShop but not found in the database:",
        'sync_success'        => "All products are in Sync!",
        'confirm_success'     => "Order status update complete, the following %s order(s) were updated: %s",
        'err_no_category'     => "Could not find category title, the category must be added at the cart level admin before this product can be uploaded!",
        'msg_sku_missing'     => "The following SKUs are on the cart but are not flagged to be there by ",
        'msg_sku_sync_success'=> "All products are in Sync.",
        'msg_test_success'    => "Connected to cart.",
    );

    function __construct($dbCreds) {
        $this->trap = true;
        $this->trace  = "Bizuno trace information for debug purposes. Generated ".date('Y-m-d H:i:s')."\n\n";
        $this->trace .= "GET Vars = " .print_r($_GET,  true)."\n";
        $this->trace .= "POST Vars = ".print_r($_POST, true)."\n";
        $type = $dbCreds['type'];
        $host = $dbCreds['host'];
        $name = $dbCreds['name'];
        $user = $dbCreds['user'];
        $pass = $dbCreds['pass'];
        $this->PDO_db = new bizunoPDO($type, "$type:host=$host;dbname=$name", $user, $pass);
    }

    // Message methods
    function msgAdd($message, $level='error') {
        $this->arrMsg[$level][] = array('text' => $message);
        $this->msgDebug("Writing to msgStack: $message");
    }
    function msgDebug($text) { $this->trace .= "\n" . (is_array($text) ? print_r($text, true) : $text); }
    function msgDebugWrite() {
        if (!$this->trap || strlen($this->trace) < 1) { return; }
        $this->trace .= "\n\nMsg array contains: ".print_r($this->arrMsg, true);
        if ($this->fileWrite($this->trace, dirname(__FILE__).'/bizDebug.txt')) { unset($this->trace); }
    }
    function msgOutput() {
        if (DEBUG) { $this->msgDebugWrite(); }
        die (json_encode($this->arrMsg));
    }

    /****************************************** Database *************************************************/
    /**
     * This method updates/inserts data into a table
     * @param string $table - string table name
     * @param array $data - example array("field_name" => 'value_update' ...)
     * @param string $action - insert or update
     * @param string $parameters
     * @return record id for insert, affected rows for update, false on error
     */
    function dbWrite($table, $data, $action = 'insert', $parameters = '', $quote = true) {
        if (! is_array ( $data ) || sizeof ( $data ) == 0) { return false; }
        $columns = array();
        if ($action == 'insert') {
            $query = "INSERT INTO $table (`" . implode ( '`, `', array_keys ( $data ) ) . "`) VALUES (";
            foreach ($data as $value) {
                switch((string)$value) {
                    case 'now()': $query .= "now(), "; break;
                    case 'null':  $query .= "null, ";  break;
                    default:      $query .= $this->PDO_db->quote($value).", "; break;
                }
            }
            $query = substr ( $query, 0, - 2 ) . ')';
        } elseif ($action == 'update') {
            foreach ($data as $column => $value) {
                switch ((string)$value) {
                    case 'now()' : $columns [] = "`$column`=NOW()"; break;
                    case 'null' :  $columns [] = "`$column`=NULL";  break;
                    default :      $columns [] = $quote ? "`$column`=".$this->PDO_db->quote($value) : "`$column`=$value"; break;
                }
            }
            $query = "UPDATE $table SET ".implode(', ', $columns).($parameters != '' ? " WHERE $parameters" : '');
        }
        $this->msgDebug("\nExecuting dbWrite query: ".$query);
        if (!$result = $this->PDO_db->Execute($query, $action)) {
            if ($this->PDO_db->msg) { $this->msgDebug($this->PDO_db->msg); }
            if ($this->PDO_db->msg) { $this->msgAdd($this->PDO_db->msg); }
        }
        return $result;
    }

    /**
     * This function pulls a value or array of values from a db table
     * @param string $table - database table name
     * @param mixed $field - one (string) or more (array) fields to retrieve
     * @param string $filter - criteria to limit results
     * @param boolean $quote - add quotes to field names, leave off for executing sql functions, i.e. SUM(field)
     * @return boolean|multitype:unknown if field is string, keyed array if field is array
     */
    function dbGetValue($table, $field, $filter = false, $quote = true) {
        $is_array = is_array ( $field ) ? true : false;
        if (! is_array ( $field )) { $field = [$field]; }
        $sql = "SELECT ".($quote ? ("`".implode ('`, `', $field )."`") : implode(', ', $field ))."
            FROM $table".($filter ? " WHERE $filter" : '')." LIMIT 1";
        $this->msgDebug("\nExecuting dbGetValue query: ".$sql);
        if (!$result = $this->PDO_db->Execute($sql, 'row')) {
            if ($this->PDO_db->msg) { $this->msgDebug($this->PDO_db->msg); }
            if ($this->PDO_db->msg) { $this->msgAdd($this->PDO_db->msg); }
            return false;
        }
        if ($is_array) { return $result; }
        return is_array($result) ? array_shift($result) : false;
    }

    /**
     * This function pulls a single row froma db table
     * @param string $table
     * @param string $filter
     * @return array - table row results, false if error/no data
     */
    function dbGetRow($table, $filter = '') {
        $sql = "SELECT * FROM `$table`".($filter ? " WHERE $filter" : '')." LIMIT 1";
        $this->msgDebug("\nExecuting dbGetRow query: ".$sql);
        if (!$result = $this->PDO_db->Execute($sql, 'row')) {
            if ($this->PDO_db->msg) { $this->msgDebug($this->PDO_db->msg); }
            if ($this->PDO_db->msg) { $this->msgAdd($this->PDO_db->msg); }
        }
        return $result;
    }

    /**
     * This function pulls multiple rows from a db table
     * @param string $table - db table name
     * @param string $filter - criteria to limit data
     * @param string $order - sort order of result
     * @return array - empty for no hits or array or rows (keyed array) for one or more hits
     */
    function dbGetMulti($table, $filter = '', $order = '', $field='*') {
        if (is_array($field)) { $field = "`".implode('`, `', $field)."`"; }
        elseif ($field != "*") { $field = "`$field`"; }
        $sql = "SELECT $field FROM $table".($filter ? " WHERE $filter" : '').($order ? " ORDER BY $order" : '');
        $this->msgDebug("\nExecuting dbGetMulti query: ".$sql);
        if (!$result = $this->PDO_db->Execute($sql, 'rows')) {
            if ($this->PDO_db->msg) { $this->msgDebug($this->PDO_db->msg); }
            if ($this->PDO_db->msg) { $this->msgAdd($this->PDO_db->msg); }
        }
        return $result;
    }

    /**
     * This method executes a query and returns the resulting PDO statement
     * @param string $sql - the QUOTED, ESCAPED sql to be executed
     */
    function dbGetResult($sql) {
        $this->msgDebug("\nExecuting dbGetResult query: ".$sql);
        if (!$result = $this->PDO_db->Execute($sql)) {
            if ($this->PDO_db->msg) { $this->msgDebug($this->PDO_db->msg); }
            if ($this->PDO_db->msg) { $this->msgAdd($this->PDO_db->msg); }
        }
        return $result;
    }

    private function fileWrite($data, $filename, $verbose=true) {
        if (strlen($data) < 1) { return false; }
        $path = substr($filename, 0, strrpos($filename, '/') + 1); // pull the path from the full path and file
        if (!is_dir($path)) { mkdir($path,0755,true); }
        if (!$handle = @fopen($filename, 'w')) {
            if ($verbose) { $this->msgAdd(sprintf($this->lang('err_io_file_open'), $filename)); }
            return false;
        }
        if (@fwrite($handle, $data) === false) {
            if ($verbose) { $this->msgAdd(sprintf($this->lang('err_io_file_write'), $filename)); }
            return false;
        }
        @fclose($handle);
        return true;
    }

    /**
     * Searches for the country codes/name and returns codes/name
     * @param string $country The search value
     * @param string $format the format of the search value, default is iso2. Acceptable values: iso2, iso3, name.
     * @param string $return the format of the return value, default is iso3. Acceptable values: iso2, iso3, name.
     * @return string/boolean the desired code/name or country on failure
     */
    public function get_country_iso($country, $format='ISO2', $return='ISO3') {
        $tmp    = file_get_contents($this->path_to_this_module.'locale.xml');
        $result = simplexml_load_string(trim($tmp), 'SimpleXMLElement', LIBXML_NOCDATA);
        $locales= json_decode(json_encode($result)); // convert from simpleXMLObject to stdClass
        if (!$locales) { $this->msgDebug("\nFailed parsing country xml"); return $country; } // if library not loaded, just return
        foreach ($locales->Locale as $locale) { if ($locale->Country->$format == $country) {
            return $locale->Country->$return;
        } }
        return $country;
    }
}

/** --------------------------------------------------------------------------------------
 * Generic PDO class wrapper
 */
class bizunoPDO extends \PDO {
    var $total_count = 0;
    var $total_time = 0;
    public $connected = false;
    private $PDO_statment;
    private $db_options;

    /**
     * This is tested to mysql only
     * @param unknown_type $driver
     * @param unknown_type $dns
     * @param unknown_type $user
     * @param unknown_type $password
     */
    function __construct($driver, $dns, $user, $password) {
        $this->driver = $driver;
        switch ($driver) {
            default :
            case "mysql" :
                try {
                    parent::__construct ( $dns, $user, $password );
                } catch ( Exception $e ) {
                    return false; // msg not in this class$this->msgAdd ( 'Connection failed: ' . $e->getMessage (), 'warning' );
                }
                $this->setAttribute(\PDO::MYSQL_ATTR_USE_BUFFERED_QUERY, true );
                $this->db_options   = array(\PDO::MYSQL_ATTR_USE_BUFFERED_QUERY, true);
                $this->PDO_statment = $this->exec("SET character_set_results='utf8', character_set_client='utf8', character_set_connection='utf8', character_set_database='utf8', character_set_server='utf8'");
                break;
        }
        $this->connected = true;
    }

    /**
     *
     * @param unknown_type $sql
     * @param unknown_type $data
     * @return boolean|PDOStatement
     */
    function db_execute_sql($sql, $data) {
        $statment = $this->prepare ( $sql, $this->db_options );
        if ($statment->execute ( $data ) === false) {
            echo "unable execute: $sql <br> data is: ";
            print_r ( $data );
            echo " <br>";
            return false;
        }
        return $statment;
    }

    /**
     * Generic sql query wrapper for error logging and messages
     * @param string $sql
     * @param string $action
     *            - action to perform
     * @return false on error, array or statement on success
     */
    function Execute($sql, $action = 'stmt') {
        if (!$this->connected) { return $this->msg = 'ERROR: Not connected to the db!'; }
        $error  = false;
        $output = false;
        $this->msg = '';
        switch ($action) {
            case 'insert' : // returns id of new row inserted
                if (false !== $this->exec ( $sql )) { $output = $this->lastInsertId(); }
                else { $error = true; }
                break;
            case 'update' : // returns affected rows
                $output = $this->exec($sql);
                if ($output === false) { $error = true; }
                break;
            case 'row' : // returns single table row
                $stmt = $this->query($sql);
                if ($stmt) { $output = $stmt->fetch(\PDO::FETCH_ASSOC); }
                else { $error = true; }
                break;
            case 'rows' : // returns array of one or more table rows
                $stmt = $this->query($sql);
                if ($stmt) { $output = $stmt->fetchAll(\PDO::FETCH_ASSOC); }
                else { $error = true; }
                break;
            default :
            case 'stmt' : // PDO Statement
                if (!$output = $this->query($sql)) { $error = true; }
                break;
        }
        if ($error) {
            $this->msg = "SQL Error: action: $action SQL: $sql returned error:".print_r($this->errorInfo(), true);
            return false;
        }
        return $output;
    }
}
